/*
 * wifi.c
 *
 *  Created on: 07.07.2014
 *      Author: AndreR
 */

#include "wifi.h"

#include "util/console.h"
#include "util/bits.h"
#include "util/sys_time.h"
#include "util/init_hal.h"
#include "util/log.h"
#include "util/angle_math.h"
#include "constants.h"
#include "commands.h"
#include "errors.h"
#include "network.h"

#include <string.h>
#include <math.h>

WifiGlobal wifi = {
	.cfg = { 100, 2, 8, 0, 1000 },
};

void EXTI2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = GPIO_PIN_2;

	NRF24ExIRQ(&wifi.nrf24);

	CH_IRQ_EPILOGUE();
}

void DMA1_Stream0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	SPISyncDmaRxInterrupt(&wifi.nrf24.spi);

	CH_IRQ_EPILOGUE();
}

void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;
	TIM7->CR1 = 0;

	chSysLockFromISR();
	chBSemSignalI(&wifi.timerSem);
	chSysUnlockFromISR();

	CH_IRQ_EPILOGUE();
}

static void startWifiTimeout(uint32_t microSeconds)
{
	TIM7->ARR = microSeconds;
	TIM7->CR1 = TIM_CR1_CEN | TIM_CR1_OPM;
}

static void outgoingCmdCallback(uint8_t id, uint8_t* pData, uint32_t pktSize)
{
	(void)pktSize;

	PacketHeader* pHeader = (PacketHeader*)pData;
	if(pHeader->section & SECTION_EXTENDED_PACKET)
		pData += PACKET_HEADER_EX_SIZE;
	else
		pData += PACKET_HEADER_SIZE;

	WifiRobot* pBot = &wifi.bots[id];

	if(pHeader->section == SECTION_SYSTEM && pHeader->cmd == CMD_SYSTEM_MATCH_CTRL)
	{
		SystemMatchCtrl* pCtrl = (SystemMatchCtrl*)pData;

		memcpy(pCtrl->curPosition, pBot->lastVisionPos, 3*sizeof(int16_t));
		uint32_t delay = (SysTimeUSec() - pBot->tVisionCapture)/250;
		if(delay > 255)
			delay = 255;
		pCtrl->posDelay = delay;
		pCtrl->camId = pBot->lastVisionCamId;
		if(network.visionAvailable == 0)
			pCtrl->camId |= SYSTEM_MATCH_CTRL_CAM_ID_FLAG_NO_VISION;

		pBot->lastVisionPos[0] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		pBot->lastVisionPos[1] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		pBot->lastVisionPos[2] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
	}
}

void WifiInit()
{
	int16_t result;
	GPIOInitData gpioInit;

	chBSemObjectInit(&wifi.timerSem, TRUE);
	chMtxObjectInit(&wifi.nrf24Mutex);

	FlashFSOpenOrCreate("wifi/cfg", 2, &wifi.cfg, sizeof(WifiConfig), &wifi.pCfgFile);

	// NRF24 IRQ line
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;	// active low
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOD, GPIO_PIN_2, &gpioInit);	// IRQ line

	NVICEnableIRQ(EXTI2_IRQn, IRQL_WIFI);

	// SPI3
	RCC->APB1ENR |= RCC_APB1ENR_SPI3EN;

	NVICEnableIRQ(DMA1_Stream0_IRQn, IRQL_SPI3);

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 6;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_50MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOC, GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit);	// SCK, MISO, MOSI

	// TIM7 setup
	RCC->APB1ENR |= RCC_APB1ENR_TIM7EN;

	TIM7->CR1 = 0;
	TIM7->CR2 = 0;
	TIM7->SR = 0;
	TIM7->DIER = TIM_DIER_UIE;
	TIM7->PSC = 83; 	// clock is now 1MHz
	TIM7->ARR = 1000;	// = 1ms

	NVICEnableIRQ(TIM7_IRQn, IRQL_NRF24_TIMER);

	for(uint8_t i = 0; i <= NRF24_MAX_BOT_ID; i++)
	{
		WifiRobot* pBot = &wifi.bots[i];

		NRF24QueueInit(&pBot->queue, pBot->txData, WIFI_TX_SIZE, pBot->rxData, WIFI_RX_SIZE);

		wifi.bots[i].lastResponseTime = -2*wifi.cfg.timeout;
		wifi.bots[i].online = 0;
		wifi.bots[i].queue.id = i;
		wifi.bots[i].lastVisionPos[0] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		wifi.bots[i].lastVisionPos[1] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		wifi.bots[i].lastVisionPos[2] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;

		NRF24QueueSetOutgoingCmdCallback(&pBot->queue, &outgoingCmdCallback);
	}

	// NRF24 setup
	NRF24ExData data;

	data.csPin = GPIO_PIN_7;
	data.pCSPort = GPIOD;
	data.spiData.pRegister = SPI3;
	data.spiData.pDma = DMA1;
	data.spiData.dmaChRx = 0;
	data.spiData.dmaChTx = 7;
	data.rx = 0;
	data.channel = wifi.cfg.channel;
	data.speed = wifi.cfg.speed;
	data.pCEPort = GPIOD;
	data.cePin = GPIO_PIN_3;

	result = NRF24ExInit(&wifi.nrf24, &data);
	if(result)
		ConsolePrint("NRF24Init failed: 0x%02hX\r\n", result);
}

void WifiSendPose(uint16_t botId, float x, float y, float orient, uint32_t tCapture, uint8_t camId)
{
	if(botId > NRF24_MAX_BOT_ID)
	{
		ConsolePrint("Invalid pose id %hu\r\n", botId);
		return;
	}

	WifiRobot* pBot = &wifi.bots[botId];

	if(pBot->lastVisionCamId != camId && chVTTimeElapsedSinceX(pBot->visionCamRxTime) < MS2ST(50))
	{
		return;
	}

	const float maxVelXY = 5.0f;
	const float maxVelW = 50.0f;
	float visionDt = (tCapture-pBot->tVisionCapture)*1e-6f;

	float searchRadiusXY = maxVelXY*visionDt;
	float searchRadiusW = maxVelW*visionDt;

	float diffX = (x - pBot->lastVisionPosFloat[0])*1e-3f;
	float diffY = (y - pBot->lastVisionPosFloat[1])*1e-3f;
	float diffXY = sqrtf(diffX*diffX+diffY*diffY);
	float diffW = AngleNormalize(orient - pBot->lastVisionPosFloat[2]);

	if(diffXY > searchRadiusXY || fabsf(diffW) > searchRadiusW)
	{
		// this is an invalid sample, vel is impossible
		return;
	}

	pBot->lastVisionPosFloat[0] = x;
	pBot->lastVisionPosFloat[1] = y;
	pBot->lastVisionPosFloat[2] = orient;

	pBot->tVisionCapture = tCapture;
	pBot->lastVisionPos[0] = (int16_t)(x+0.5f);
	pBot->lastVisionPos[1] = (int16_t)(y+0.5f);
	pBot->lastVisionPos[2] = (int16_t)(orient*1000.0f+0.5f);
	pBot->lastVisionCamId = camId;
	pBot->visionCamRxTime = chVTGetSystemTimeX();
}

void WifiSaveConfig()
{
	FlashFSWrite(&wifi.pCfgFile, &wifi.cfg, sizeof(WifiConfig));
}

void WifiSetChannel(uint8_t channel)
{
	chMtxLock(&wifi.nrf24Mutex);
	NRF24ExChangeChannel(&wifi.nrf24, channel);
	chMtxUnlock(&wifi.nrf24Mutex);

	wifi.cfg.channel = channel;
}

void WifiSetSpeed(uint8_t speed)
{
	chMtxLock(&wifi.nrf24Mutex);
	int16_t result = NRF24ExChangeSpeed(&wifi.nrf24, speed);
	chMtxUnlock(&wifi.nrf24Mutex);

	if(result)
		return;

	wifi.cfg.speed = speed;
}

void WifiSetMaxBot(uint8_t maxBots)
{
	if(maxBots > NRF24_MAX_BOT_ID+1)
		return;

	wifi.cfg.maxBots = maxBots;
}

void WifiSetFixedRuntime(uint8_t enable)
{
	wifi.cfg.fixedRuntime = enable;
}

void WifiSetTimeout(uint32_t ms)
{
	wifi.cfg.timeout = ms;
}

void WifiSetBotDataCallback(BotDataAvailCallback cb)
{
	wifi.cbBotData = cb;
}

void WifiPrintConfig()
{
	ConsolePrint("--- Wireless Configuration ---\r\n");
	ConsolePrint("Channel:   %hu (2.%03huGHz)\r\n", (uint16_t)wifi.cfg.channel, ((uint16_t)wifi.cfg.channel)+400);

	if(wifi.cfg.speed == 0)
		ConsolePrint("Speed:     250k\r\n");
	if(wifi.cfg.speed == 1)
		ConsolePrint("Speed:     1M\r\n");
	if(wifi.cfg.speed == 2)
		ConsolePrint("Speed:     2M\r\n");

	ConsolePrint("Max. Bots: %hu\r\n", (uint16_t)wifi.cfg.maxBots);
	ConsolePrint("Timeout:   %ums\r\n", wifi.cfg.timeout);

	if(wifi.cfg.fixedRuntime)
		ConsolePrint("Runtime:   fixed\r\n");
	else
		ConsolePrint("Runtime:   variable\r\n");
}

static int16_t processBot(WifiRobot* pBot)
{
	int16_t result;

	if(chVTTimeElapsedSinceX(pBot->lastTimeSync) > S2ST(5) && sysTime.unixOffset != 0)
	{
		pBot->lastTimeSync = chVTGetSystemTimeX();

		uint8_t data[sizeof(PacketHeader)+sizeof(uint32_t)];

		PacketHeader* pHeader = (PacketHeader*)data;
		pHeader->section = SECTION_SYSTEM;
		pHeader->cmd = CMD_SYSTEM_TIMESYNC;

		uint32_t* pTime = (uint32_t*)(data+sizeof(PacketHeader));
		*pTime = SysTimeUnix();

		NRF24QueueSend(&pBot->queue, data, sizeof(PacketHeader)+sizeof(uint32_t));
	}

	chBSemReset(&wifi.timerSem, 1);

	switch(wifi.cfg.speed)
	{
		case 0: startWifiTimeout(3494); break;
		case 1: startWifiTimeout(1194); break;
		default: startWifiTimeout(994); break;
	}

	LogDebugC("NRF24ExUseQueue", pBot->queue.id);
	NRF24ExUseQueue(&wifi.nrf24, &pBot->queue);

	LogDebugC("NRF24ExTxProcessOne", pBot->queue.id);
	uint16_t rxPkt = pBot->queue.stats.nrfIO.rxPackets;
	result = NRF24ExTxProcessOne(&wifi.nrf24);
	if(result != 0 && result != ERROR_NRF24EX_TIMEOUT)
	{
		ConsolePrint("NRF24ExProcess error: 0x%0X\r\n", result);

		NRF24ExSetup(&wifi.nrf24);
		NRF24ExUseQueue(&wifi.nrf24, &pBot->queue);
	}

	if(rxPkt != pBot->queue.stats.nrfIO.rxPackets)
		pBot->lastResponseTime = chVTGetSystemTime();

	LogDebugC("CB", pBot->queue.id);
	if(pBot->queue.rxFifo.numPackets && wifi.cbBotData)
		(*wifi.cbBotData)(pBot->queue.id);

	LogDebugC("ConsolePrint1", pBot->queue.id);
	// check if bot timed out
	uint32_t elapsedTime = chVTTimeElapsedSinceX(pBot->lastResponseTime);
	if(elapsedTime > MS2ST(wifi.cfg.timeout))
	{
		// no response => timeout
		if(pBot->online)
		{
			pBot->online = 0;
			NRF24QueueClear(&pBot->queue);

			ConsolePrint("Bot %hu timed out after %u ms. Elapsed: %u\r\n", (uint16_t)pBot->queue.id, wifi.cfg.timeout, elapsedTime);
		}
	}
	else
	{
//		pBot->lastResponseTime = chVTGetSystemTime()-(2*wifi.cfg.timeout);

		if(!pBot->online)
		{
			pBot->online = 1;

			ConsolePrint("Bot %hu online\r\n", (uint16_t)pBot->queue.id);
		}
	}

	LogDebugC("chBSemWait", pBot->queue.id);
	chBSemWait(&wifi.timerSem);

	LogDebugC("Done", pBot->queue.id);

	return result;
}

void WifiTask(void* params)
{
	(void) params;

	chRegSetThreadName("Wifi");

	ConsolePrint("WifiTask started\r\n");

	uint32_t startTime = SysTimeUSec();
	uint32_t endTime;
	uint8_t checkId = 0;

	chMtxLock(&wifi.nrf24Mutex);
	NRF24ExSetup(&wifi.nrf24);
	NRF24ExUseQueue(&wifi.nrf24, &wifi.bots[0].queue);
	chMtxUnlock(&wifi.nrf24Mutex);

	while(1)
	{
		// blocking here, probably a race condition/wrong unlock/lock before
		chMtxLock(&wifi.nrf24Mutex);
		startTime = SysTimeUSec();

		uint8_t numBotsOnline = 0;
		for(uint8_t i = 0; i <= NRF24_MAX_BOT_ID; i++)
		{
			WifiRobot* pBot = &wifi.bots[i];

			if(!pBot->online)
				continue;

			++numBotsOnline;

			processBot(pBot);
		}

		while(numBotsOnline < wifi.cfg.maxBots)
		{
			do
			{
				++checkId;
				checkId %= NRF24_MAX_BOT_ID+1;
			}
			while(wifi.bots[checkId].online);

			processBot(&wifi.bots[checkId]);

			++numBotsOnline;

			if(wifi.cfg.fixedRuntime == 0)	// only one check?
				break;
		}

		endTime = SysTimeUSec();

		chMtxUnlock(&wifi.nrf24Mutex);

		wifi.runtime = endTime-startTime;

		wifi.lastCall = SysTimeUSec();
	}
}
