/*
 * motor.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

/**	Timer/ADC/Motor assignments
 *
 * 			Dribbler:
 * 	TIM1:	M5
 * 	ADC3:	M5B
 * 	ADC4:	M5A
 */

#pragma GCC optimize ("Ofast")

#include "motor.h"
#include "arm_math.h"

#include "../constants.h"
#include "ch.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include "util/console.h"
#include "util/init_hal.h"
#include "../supply.h"
#include <math.h>

#define PWM_MAX_VAL 3599
#define PI_3RD 1.047197551196598f
//#define MOT_ADC_TO_CURRENT 0.00244140625f
//#define MOT_ADC_TO_CURRENT 0.0029296875f
#define MOT_ADC_TO_CURRENT 0.00341796875f
#define MOTOR_RESISTANCE 0.2015f	// [Ohm]
#define H5Mask ((1 << 13) | (1 << 14) | (1 << 15))
#define M5Pos ((GPIOC->IDR & H5Mask) >> 13)

#define CCMRx_OCxPE (TIM_CCMR1_OC1PE | TIM_CCMR1_OC2PE)
#define CCMR1_OC1_PWM2 (7 << 4)
#define CCMR1_OC1_LOW_ON (5 << 4)
#define CCMR1_OC2_PWM2 (7 << 12)
#define CCMR1_OC2_LOW_ON (5 << 12)
#define CCMR2_OC3_PWM2 (7 << 4)
#define CCMR2_OC3_LOW_ON (5 << 4)

/**	pos	W1	W2	W3
 *	0	-	-	-
 *	1	+	.	-
 *	2	-	+	.
 *	3	.	+	-
 *	4	.	-	+
 *	5	+	-	.
 *	6	-	.	+
 *	7	.	.	.
 */

// W3  W2  W1
//  5   5   5
static const uint32_t ccerLUT[8] = {
	0x555,
	0x505,
	0x055,
	0x550,
	0x550,
	0x055,
	0x505,
	0x000
};

static const uint32_t ccmr1LUT[8] = {
	CCMR1_OC1_LOW_ON | CCMR1_OC2_LOW_ON | CCMRx_OCxPE,
	CCMR1_OC1_PWM2 | CCMRx_OCxPE,
	CCMR1_OC1_LOW_ON | CCMR1_OC2_PWM2 | CCMRx_OCxPE,
	CCMR1_OC2_PWM2 | CCMRx_OCxPE,
	CCMR1_OC2_LOW_ON | CCMRx_OCxPE,
	CCMR1_OC1_PWM2 | CCMR1_OC2_LOW_ON | CCMRx_OCxPE,
	CCMR1_OC1_LOW_ON | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
};

static const uint32_t ccmr2LUT[8] = {
	CCMR2_OC3_LOW_ON | CCMRx_OCxPE,
	CCMR2_OC3_LOW_ON | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
	CCMR2_OC3_LOW_ON | CCMRx_OCxPE,
	CCMR2_OC3_PWM2 | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
	CCMR2_OC3_PWM2 | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
};

Motor motor;

Motor* pMotor = &motor;

static float min3(float a, float b, float c);
static inline void updateADC();
static inline void updateHall(uint16_t hallPos);

IN_CCM_VAR static uint32_t lastCOMTime;

/**
 * IRQs
 */
IN_CCM void TIM1_UP_TIM16_IRQHandler()
{
	TIM1->SR = 0;

	updateADC();
}

// 100Hz speed update
IN_CCM void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;

	uint32_t now = TIM2->CNT;

	if(now-lastCOMTime > 10000000)
	{
		lastCOMTime = now;
		motor.curVelocity = 0;
	}

	PIDUpdate(&motor.pidSpeed, pMotor->curVelocity);

	float cur = motor.rmsCurrent;
	if(motor.pidSpeed.setpoint < 100.0f)
	{
		EMAFilterUpdate(&motor.biasCurrent, cur);
	}

	cur -= motor.biasCurrent.value;

	motor.temp = LagElementPT1Process(&motor.tempPT1, cur)+25.0f;

	if(motor.temp > 90.0f)
	{
		motor.overheated = 1;
	}

	if(motor.overheated)
	{
		MotorSpeedSetVelocity(0);
	}

	if(motor.temp < 70.0f)
	{
		motor.overheated = 0;
	}

	if(motor.pidSpeed.enabled)
	{
		float setVoltage = motor.pidSpeed.output*MOTOR_KE;

		if(motor.pidSpeed.setpoint < 1000)
		{
			setVoltage = 0;
			motor.pidSpeed.iTerm = 0;
		}

		MotorSetVoltage(setVoltage);
	}

	CH_IRQ_EPILOGUE();
}

IN_CCM void EXTI15_10_IRQHandler()
{
	IN_CCM_VAR static uint8_t lastPos;

	EXTI->PR = EXTI_PR_PR13 | EXTI_PR_PR14 | EXTI_PR_PR15;

	// commutation logic
	uint8_t halPos1 = M5Pos;
	for(uint8_t i = 0; i < 10; i++)
		asm volatile("nop");

	uint8_t halPos2 = M5Pos;
	if(halPos1 != halPos2 || halPos2 == lastPos)
		return; // spike and change filter

	updateHall(halPos1);

	lastPos = halPos1;

	// delta time measurement between hall events
	uint32_t now = TIM2->CNT;
	uint32_t dT = now-lastCOMTime;
	lastCOMTime = now;

	float rpm = 720e6f/((float)dT);

	const float alpha = 0.8f;
	motor.curVelocity = (alpha*motor.curVelocity + (1.0f-alpha)*rpm);
}

static void initHw()
{
	GPIOInitData gpioInit;

	// ### DMA configuration (ADC)
	// peripheral-to-memory, half-words, medium priority, memory increment, circular mode
	DMA2_Channel5->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel5->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel5->CPAR = (uint32_t)&ADC3->DR;
	DMA2_Channel5->CMAR = (uint32_t)pMotor->adc[1];

	DMA2_Channel2->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel2->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel2->CPAR = (uint32_t)&ADC4->DR;
	DMA2_Channel2->CMAR = (uint32_t)pMotor->adc[0];

	// ### ADC
	RCC->AHBENR |= RCC_AHBENR_ADC34EN;

	gpioInit.mode = GPIO_MODE_ANALOG;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOB, GPIO_PIN_13 | GPIO_PIN_14, &gpioInit); // B2, B13, B14

	// independent, synchronous clock from AHB
	ADC3_4->CCR = ADC34_CCR_CKMODE_0;

	// enable voltage regulator for each ADC
	ADC3->CR = ADC_CR_ADVREGEN_0;
	ADC4->CR = ADC_CR_ADVREGEN_0;
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	// start single-ended calibration
	ADC3->CR |= ADC_CR_ADCAL;
	ADC4->CR |= ADC_CR_ADCAL;
	chThdSleep(MS2ST(4));	// calibration delay

	pMotor->adcCal[0]  = ADC3->CALFACT;
	pMotor->adcCal[1]  = ADC4->CALFACT;

	// continuous conversion, 12Bit, trigger TIM1_TRGO2
	ADC3->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC3->SQR1 = DRIVE_ADC_SAMPLES-1;

	ADC4->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC4->SQR1 = DRIVE_ADC_SAMPLES-1;

	for(uint8_t i = 1; i <= DRIVE_ADC_SAMPLES; i++)
	{
		ADCRegularChannelConfig(ADC3, 5,  i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC4, 4,  i, ADC_SampleTime_181Cycles5);
	}

	// configure offsets
	ADC3->OFR1 = ADC_OFR1_OFFSET1_EN | ( 5 << 26) | 1706;
	ADC4->OFR1 = ADC_OFR1_OFFSET1_EN | ( 4 << 26) | 1706;

	// Enable the ADC DMA
	ADC3->CFGR |= ADC_CFGR_DMAEN;
	ADC4->CFGR |= ADC_CFGR_DMAEN;

	ADC3->CR |= ADC_CR_ADEN;
	ADC4->CR |= ADC_CR_ADEN;
	chThdSleep(MS2ST(4));	// ADC ready delay

	// Enable the DMA channel
	DMA2_Channel5->CCR |= DMA_CCR_EN;
	DMA2_Channel2->CCR |= DMA_CCR_EN;

	// ### Motors
	RCC->APB2ENR |= RCC_APB2ENR_TIM1EN;
	RCC->CFGR3 |= RCC_CFGR3_TIM1SW;	// TIM1 + TIM8 clocks now running at 144MHz

	// Timer config
	// TIM1
	TIM1->CR1 = TIM_CR1_CMS_0;	// center-aligned mode 1
	TIM1->PSC = 0;
	TIM1->ARR = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	TIM1->RCR = 1;				// update event only every second count (e.g. downcounting) = 20kHz

	// PWM mode 2, preload enable on CCR(1-3) registers
	TIM1->CCMR1 = TIM_CCMR1_OC1PE | (7 << 4) | TIM_CCMR1_OC2PE | (7 << 12);
	TIM1->CCMR2 = TIM_CCMR2_OC3PE | (7 << 4) | (7 << 12);
	TIM1->CCER = TIM_CCER_CC1E | TIM_CCER_CC1NE | TIM_CCER_CC2E | TIM_CCER_CC2NE | TIM_CCER_CC3E | TIM_CCER_CC3NE | TIM_CCER_CC4E;
	TIM1->CR2 = TIM_CR2_OIS1 | TIM_CR2_OIS2 | TIM_CR2_OIS3 | (7 << 20);	// output idle state = high, OC4 on TRGO2 (connected to ADC)
	TIM1->CCR1 = 0;
	TIM1->CCR2 = 0;
	TIM1->CCR3 = 0;
	TIM1->CCR4 = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM1->EGR = TIM_EGR_UG;

	TIM1->BDTR = TIM_BDTR_AOE | TIM_BDTR_BKP | TIM_BDTR_OSSR;
	TIM1->DIER = TIM_DIER_UIE;

	TIM1->CR1 |= TIM_CR1_CEN;

	ADC3->CR |= ADC_CR_ADSTART;
	ADC4->CR |= ADC_CR_ADSTART;

	TIM1->BDTR |= TIM_BDTR_MOE;

	// pin config - drive outputs
	GPIOA->ODR |= GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12;
	GPIOB->ODR |= GPIO_PIN_15;

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 6;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_2MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit);

	gpioInit.alternate = 4;
	GPIOInit(GPIOB, GPIO_PIN_15, &gpioInit);

	// ### Hall Sensors
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING_FALLING;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOC, GPIO_PIN_13 | GPIO_PIN_14 | GPIO_PIN_15, &gpioInit);

	// Control timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM7EN;
	TIM7->CR1 = 0;
	TIM7->CR2 = 0;
	TIM7->PSC = 719;	// = 10us
	TIM7->ARR = 999;	// = 10ms = 100Hz
	TIM7->RCR = 0;
	TIM7->DIER = TIM_DIER_UIE;
	TIM7->CR1 |= TIM_CR1_CEN;

	// Free running time measurement timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;
	TIM2->CR1 = 0;
	TIM2->CR2 = 0;
	TIM2->PSC = 0;
	TIM2->ARR = 0xFFFFFFFF;
	TIM2->RCR = 0;
	TIM2->CR1 |= TIM_CR1_CEN;

	NVICEnableIRQ(TIM1_UP_TIM16_IRQn, 3);
	NVICEnableIRQ(EXTI15_10_IRQn, 2);
	NVICEnableIRQ(TIM7_IRQn, IRQL_MOTOR_UPDATE+2);
}

/**
 * PRIVATE functions
 */
IN_CCM static float min3(float a, float b, float c)
{
	if(a < b && a < c)
		return a;

	if(b < a && b < c)
		return b;

	return c;
}

typedef union _i32i16
{
	int32_t i32;
	int16_t i16[2];
} i32i16_t;

IN_CCM static inline void updateADC()
{
	int16_t i016, i116;
	float32_t i0f, i1f;
	i32i16_t i0i, i1i;

	i0i.i32 = motor.adc32[0][0];
	i1i.i32 = motor.adc32[1][0];

	for(uint8_t i = 1; i < DRIVE_ADC_SAMPLES/2; i++)
	{
		i0i.i32 = __QADD16(i0i.i32, motor.adc32[0][i]);
		i1i.i32 = __QADD16(i1i.i32, motor.adc32[1][i]);
	}

	i016 = i0i.i16[0]+i0i.i16[1];
	i116 = i1i.i16[0]+i1i.i16[1];

	i0f = i016*(1.0f/DRIVE_ADC_SAMPLES);
	i1f = i116*(1.0f/DRIVE_ADC_SAMPLES);

	i0f = i0f*-MOT_ADC_TO_CURRENT; // convert to real current
	i1f = i1f*-MOT_ADC_TO_CURRENT; // convert to real current

	float a = i0f;
	float b = 0.57735026918963f*i0f + 1.154700538379252f*i1f;

	const float alpha = 0.995f;
	float rms = sqrtf(a*a + b*b);
	motor.rmsCurrent = (alpha*motor.rmsCurrent + (1.0f-alpha)*rms);
}

IN_CCM static inline void updateHall(uint16_t hallPos)
{
	motor.halPos = hallPos;

	if(motor.mode != MOTOR_MODE_BLOCK)
		return;

	if(motor.outLevel > 0)
	{
		hallPos ^= 0x7;
	}

	TIM1->CCER = ccerLUT[hallPos];

	TIM1->CCMR1 = ccmr1LUT[hallPos];
	TIM1->CCMR2 = ccmr2LUT[hallPos];
}

IN_CCM static void setCCR(uint16_t a, uint16_t b, uint16_t c)
{
	TIM1->CCR1 = a;
	TIM1->CCR2 = b;
	TIM1->CCR3 = c;
}

/**
 * PUBLIC functions
 */
void MotorInit()
{
	EMAFilterInit(&motor.biasCurrent, 0.995f);

	PIDParams speedPIDParams;
	speedPIDParams.kp = 1.0f;
	speedPIDParams.ki = 50.0f;
	speedPIDParams.kd = 0.0f;
	speedPIDParams.outputMin = 0;	// [RPM]
	speedPIDParams.outputMax = 45000.0f;	// [RPM]

	PIDInit(&motor.pidSpeed, &speedPIDParams, 0.01f, 1);

	LagElementPT1Init(&motor.tempPT1, 70, 240, 0.01f);

	initHw();

	MotorSetMode(MOTOR_MODE_BLOCK);

	MotorSetCtrlMode(MOTOR_CTRL_MODE_PID_SPEED);
}

void MotorSetMode(uint8_t mode)
{
	if(mode == motor.mode)
		return;

	motor.mode = mode;

	switch(mode)
	{
		case MOTOR_MODE_DISABLE:
		{
			TIM1->CCER = ccerLUT[7];
			motor.outLevel = 0;
		}
		break;
		case MOTOR_MODE_BLOCK:
		{
			// trigger hall IRQ to get a valid hall position
			EXTI->SWIER |= (EXTI_SWIER_SWIER10 | EXTI_SWIER_SWIER14);
		}
		break;
		case MOTOR_MODE_SVPWM:
		{
			motor.outLevel = 0;
			TIM1->CCR1 = 0;
			TIM1->CCR2 = 0;
			TIM1->CCR3 = 0;
			TIM1->CCER = 0x555;
			TIM1->CCMR1 = CCMR1_OC1_PWM2 | CCMR1_OC2_PWM2 | CCMRx_OCxPE;
			TIM1->CCMR2 = CCMR2_OC3_PWM2 | CCMRx_OCxPE;
		}
		break;
		case MOTOR_MODE_TEST1:
		{
			// W1+, W2-
			TIM1->CCER = ccerLUT[5];
			TIM1->CCMR1 = ccmr1LUT[5];
			TIM1->CCMR2 = ccmr2LUT[5];
		}
		break;
		case MOTOR_MODE_TEST2:
		{
			// W2+, W3-
			TIM1->CCER = ccerLUT[3];
			TIM1->CCMR1 = ccmr1LUT[3];
			TIM1->CCMR2 = ccmr2LUT[3];
		}
		break;
		case MOTOR_MODE_TEST3:
		{
			// W3+, W1-
			TIM1->CCER = ccerLUT[6];
			TIM1->CCMR1 = ccmr1LUT[6];
			TIM1->CCMR2 = ccmr2LUT[6];
		}
		break;
	}
}

void MotorSetCtrlMode(uint8_t mode)
{
	if(motor.ctrlMode == mode)
		return;

	switch(mode)
	{
		case MOTOR_CTRL_MODE_NONE:
		{
			PIDEnable(&motor.pidSpeed, 0);
		}
		break;
		case MOTOR_CTRL_MODE_PID_SPEED:
		{
			PIDEnable(&motor.pidSpeed, 1);
		}
		break;
		default:
			return;
	}

	motor.ctrlMode = mode;
}

IN_CCM void MotorSetVoltage(float voltage)
{
	motor.setVoltage = voltage;
	motor.setVoltageTimestamp = SysTimeUSec();

	float bat = supply.bat;
	if(bat < 9.0f || bat > 17.0f)
		bat = 11.1f;

	int32_t newOutLevel = (int32_t)(motor.setVoltage/bat*((float)PWM_MAX_VAL));
	if(newOutLevel > PWM_MAX_VAL)
		newOutLevel = PWM_MAX_VAL;

	if(newOutLevel < -PWM_MAX_VAL)
		newOutLevel = -PWM_MAX_VAL;

	motor.outLevel = newOutLevel;

	updateHall(motor.halPos);

	if(newOutLevel < 0)
		newOutLevel *= -1;

	TIM1->CCR1 = newOutLevel;
	TIM1->CCR2 = newOutLevel;
	TIM1->CCR3 = newOutLevel;
}

void MotorSpeedSetVelocity(float rpm)
{
	motor.pidSpeed.setpoint = rpm;
}

void MotorSpeedSetPIDParams(float p, float i, float d)
{
	PIDSetParams(&motor.pidSpeed, p, i, d);
}

void MotorSetPhaseVoltagesSVPWM(float Va, float Vb, float Vc)
{
	MotorSetMode(MOTOR_MODE_SVPWM);

	float Vd = supply.bat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float Vcom = -min3(Va, Vb, Vc);

	float da = 1/Vd*(Va+Vcom);
	float db = 1/Vd*(Vb+Vcom);
	float dc = 1/Vd*(Vc+Vcom);

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

void MotorSetAlphaBeta(float Valpha, float Vbeta)
{
	float Va = Valpha;
	float Vb = -0.5f*Valpha + 0.86602540378444f*Vbeta;
	float Vc = -0.5f*Valpha - 0.86602540378444f*Vbeta;

	MotorSetPhaseVoltagesSVPWM(Va, Vb, Vc);
}

void MotorSetVoltageAngle(float Vref, float angle)
{
	float sin, cos;
	while(angle > 180.0f)
		angle -= 360.0f;

	arm_sin_cos_f32(angle, &sin, &cos);

	MotorSetAlphaBeta(cos*Vref, sin*Vref);
}
