/*
 * spi4.c
 *
 *  Created on: 31.10.2015
 *      Author: AndreR
 */

#include "spi4.h"

#include "gfx.h"
#include "src/ginput/ginput_driver_mouse.h"
#include "util/init_hal.h"
#include "util/console.h"
#include "util/sys_time.h"
#include "util/log_file.h"
#include "hal/kicker.h"
#include "struct_ids.h"
#include "constants.h"
#include <string.h>

SPI4Global spi4 __attribute__((section(".dtc")));

void DMA2_Stream0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	SPISyncDmaRxInterrupt(&spi4.bus);

	CH_IRQ_EPILOGUE();
}

void SPI4Init()
{
	RCC->APB2ENR |= RCC_APB2ENR_SPI4EN;

	spi4.touch.slave.cpha = 0;
	spi4.touch.slave.cpol = 0;
	spi4.touch.slave.csPin = GPIO_PIN_4;
	spi4.touch.slave.pCSPort = GPIOG;
	spi4.touch.slave.prescaler = SPI_CR1_BRDIV64;
	spi4.touch.slave.timeoutTicks = MS2ST(20);

	SPISyncData spiData;
	spiData.dmaChRx = 0;
	spiData.dmaChTx = 1;
	spiData.dmaPrio = 1;
	spiData.pDma = DMA2;
	spiData.pRegister = SPI4;
	SPISyncInit(&spi4.bus, &spiData);
	SPISyncSlaveInit(&spi4.bus, &spi4.touch.slave);

	ADNS9800Init(&spi4.mouse[0], &spi4.bus, GPIOF, GPIO_PIN_2);
	ADNS9800Init(&spi4.mouse[1], &spi4.bus, GPIOF, GPIO_PIN_14);

	spi4.mouseLog.header.type = SID_OPTICAL_RAW;
	spi4.mouseLog.header.length = sizeof(LogOpticalRaw);

	// init GPIO
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_25MHZ;
	gpioInit.pupd = GPIO_PUPD_DOWN;
	GPIOInit(GPIOE, GPIO_PIN_2 | GPIO_PIN_5 | GPIO_PIN_6, &gpioInit); // SCK, MISO, MOSI

	// configure NVIC
	NVICEnableIRQ(DMA2_Stream0_IRQn, IRQL_SPI6);

	// init touch IRQ
	gpioInit.mode = GPIO_MODE_INPUT;
	gpioInit.alternate = 0;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOF, GPIO_PIN_12, &gpioInit);

	for(uint8_t i = 0; i < 2*SPI4_TOUCH_SAMPLES; i += 2)
		spi4.touch.tx[i] = 0x90;	// X sample

	for(uint8_t i = 2*SPI4_TOUCH_SAMPLES; i < 4*SPI4_TOUCH_SAMPLES; i += 2)
		spi4.touch.tx[i] = 0xD0;	// Y sample
}

static void updateTouch()
{
	static uint8_t inv = 0;

	SPISyncTransfer(&spi4.touch.slave, spi4.touch.tx, spi4.touch.rx, 4*SPI4_TOUCH_SAMPLES+1);

	uint16_t x = 0;
	uint16_t y = 0;

	for(uint8_t i = 1*SPI4_TOUCH_SAMPLES+1; i < 2*SPI4_TOUCH_SAMPLES+1; i += 2)
	{
		uint32_t val = ((uint16_t)spi4.touch.rx[i+1] << 5) | (spi4.touch.rx[i] >> 3);
		if(val > 4096)
			inv = 1;
		if(inv)
			val = __REV16(val);
		x += val;
	}

	for(uint8_t i = 3*SPI4_TOUCH_SAMPLES+1; i < 4*SPI4_TOUCH_SAMPLES+1; i += 2)
	{
		uint16_t val = ((uint16_t)spi4.touch.rx[i+1] << 5) | (spi4.touch.rx[i] >> 3);
		if(val > 4096)
			inv = 1;
		if(inv)
			val = __REV16(val);
		y += val;
	}

	x /= SPI4_TOUCH_SAMPLES/2;
	y /= SPI4_TOUCH_SAMPLES/2;

	spi4.touch.x = x;
	spi4.touch.y = y;

//	ConsolePrint("%hu, %hu\r\n", x, y);
}

void SPI4Task(void* params)
{
	(void)params;

	uint32_t start;
	uint32_t end;

	chRegSetThreadName("SPI4");

	ADNS9800UploadFirmware(&spi4.mouse[0]);
	ADNS9800UploadFirmware(&spi4.mouse[1]);

	systime_t prev = chVTGetSystemTimeX();
	systime_t next = prev+US2ST(10000);

	while(1)
	{
		start = SysTimeUSecRaw();
		if(spi4.mouse[0].functional)
			ADNS9800UpdateMotion(&spi4.mouse[0], 0.01f);
		if(spi4.mouse[1].functional)
			ADNS9800UpdateMotion(&spi4.mouse[1], 0.01f);
		end = SysTimeUSecRaw();
		spi4.mouseSampleTime = end-start;

		spi4.mouseLog.header.timestamp = SysTimeUSec();
		spi4.mouseLog.lDeltaX = spi4.mouse[0].motion.deltaX;
		spi4.mouseLog.lDeltaY = spi4.mouse[0].motion.deltaY;
		spi4.mouseLog.lSqual = spi4.mouse[0].motion.squal;
		spi4.mouseLog.rDeltaX = spi4.mouse[1].motion.deltaX;
		spi4.mouseLog.rDeltaY = spi4.mouse[1].motion.deltaY;
		spi4.mouseLog.rSqual = spi4.mouse[1].motion.squal;

		LogFileWrite(&spi4.mouseLog);

		// calc robot velocity
		float v[3];
		float* L = spi4.mouse[0].motion.speed;
		float* R = spi4.mouse[1].motion.speed;
		const float r = 0.07f;

		v[0] = (L[1]-R[1])/2.0f;
		v[2] = (-R[0]-L[0])/(4*r);
		v[1] = ((-v[2]*2*r-R[0])+(v[2]*2*r+L[0]))/2.0f;

		memcpy(spi4.opticalSpeed, v, sizeof(float)*3);

		if((GPIOF->IDR & GPIO_PIN_12) == 0)
		{
			spi4.touch.pressed = 1;

			start = SysTimeUSecRaw();
			updateTouch();
			end = SysTimeUSecRaw();
			spi4.touch.sampleTime = end-start;
		}
		else
		{
			spi4.touch.pressed = 0;
		}

		chThdSleepUntilWindowed(prev, next);
		prev = next;
		next += US2ST(10000);
	}
}

static bool_t mouseGetXYZ(GMouse* m, GMouseReading* pdr)
{
	(void)m;

	// No buttons
	pdr->buttons = 0;
	pdr->z = 0;

	if(spi4.touch.pressed)
	{
		pdr->z = 1;
		pdr->x = spi4.touch.x;
		pdr->y = spi4.touch.y;
	}

	return TRUE;
}

static void calSave(GMouse *m, const void *buf, size_t sz)
{
	(void)m;
	(void)sz;

	float* pCalib = (float*)buf;

	for(uint8_t i = 0; i < 6; i++)
		ConsolePrint("%.8f\r\n", pCalib[i]);
}

static const float calib[] = {
	0.00114166f, 0.06780271f, -19.17998314f,
	0.09104055f, 0.00010331f, -15.59686184f,
};

static bool_t calLoad(GMouse *m, void *buf, size_t sz)
{
	(void)m;

	if(sz != sizeof(calib))
		return FALSE;

	memcpy(buf, calib, sz);

	return TRUE;
}

static bool_t mouseInit(GMouse* m, unsigned driverinstance)
{
	(void)m;
	(void)driverinstance;

	return TRUE;
}

// Resolution and Accuracy Settings
#define GMOUSE_ADS7843_PEN_CALIBRATE_ERROR		8
#define GMOUSE_ADS7843_PEN_CLICK_ERROR			6
#define GMOUSE_ADS7843_PEN_MOVE_ERROR			4
#define GMOUSE_ADS7843_FINGER_CALIBRATE_ERROR	14
#define GMOUSE_ADS7843_FINGER_CLICK_ERROR		18
#define GMOUSE_ADS7843_FINGER_MOVE_ERROR		14

const GMouseVMT const GMOUSE_DRIVER_VMT[1] = {{
	{
		GDRIVER_TYPE_TOUCH,
		GMOUSE_VFLG_TOUCH | GMOUSE_VFLG_CALIBRATE | /*GMOUSE_VFLG_CAL_TEST |*/ GMOUSE_VFLG_ONLY_DOWN | /*GMOUSE_VFLG_POORUPDOWN |*/ GMOUSE_VFLG_CAL_EXTREMES,
		sizeof(GMouse),
		_gmouseInitDriver,
		_gmousePostInitDriver,
		_gmouseDeInitDriver
	},
	1,				// z_max - (currently?) not supported
	0,				// z_min - (currently?) not supported
	1,				// z_touchon
	0,				// z_touchoff
	{				// pen_jitter
		GMOUSE_ADS7843_PEN_CALIBRATE_ERROR,			// calibrate
		GMOUSE_ADS7843_PEN_CLICK_ERROR,				// click
		GMOUSE_ADS7843_PEN_MOVE_ERROR				// move
	},
	{				// finger_jitter
		GMOUSE_ADS7843_FINGER_CALIBRATE_ERROR,		// calibrate
		GMOUSE_ADS7843_FINGER_CLICK_ERROR,			// click
		GMOUSE_ADS7843_FINGER_MOVE_ERROR			// move
	},
	mouseInit,	 	// init
	0,				// deinit
	mouseGetXYZ,	// get
	calSave,		// calsave
	calLoad			// calload
}};
