/*
 * drive.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

/**	Timer/ADC/Motor assignments
 *
 * 			Right:	Left:
 * 	TIM1:	M1		M3
 * 	TIM2:	ENC4	ENC2
 * 	TIM4:	ENC1	ENC3
 * 	TIM8:	M4		M2
 * 	ADC1:	M4B		M2B
 * 	ADC2:	M4A		M2A
 * 	ADC3:	M1C		M3B
 * 	ADC4:	M1B		M3C
 */

#pragma GCC optimize ("Ofast")

#include "drive.h"

#include "spi1.h"
#include "intercom1.h"
#include "constants.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/arm_mat_util_f32.h"
#include "util/ema_filter.h"
#include "util/console.h"
#include <math.h>

#define PWM_MAX_VAL 3599
#define PI_3RD 1.047197551196598f
#define MOT_ADC_TO_CURRENT 0.00244140625f

Drive drive[2] = {
#ifdef MOT_RIGHT
	{ .pTIMMotor = TIM1, .pTIMEncoder = TIM4, .id = 0 },
	{ .pTIMMotor = TIM8, .pTIMEncoder = TIM2, .id = 3 },
#else
	{ .pTIMMotor = TIM8, .pTIMEncoder = TIM2, .id = 1 },
	{ .pTIMMotor = TIM1, .pTIMEncoder = TIM4, .id = 2 },
#endif
};

Drive* const pFront = &drive[0];
Drive* const pRear = &drive[1];

#define CCMRx_OCxPE (TIM_CCMR1_OC1PE | TIM_CCMR1_OC2PE)
#define CCMR1_OC1_PWM2 (7 << 4)
#define CCMR1_OC1_LOW_ON (5 << 4)
#define CCMR1_OC2_PWM2 (7 << 12)
#define CCMR1_OC2_LOW_ON (5 << 12)
#define CCMR2_OC3_PWM2 (7 << 4)
#define CCMR2_OC3_LOW_ON (5 << 4)

/**	pos	W1	W2	W3
 *	0	-	-	-
 *	1	+	.	-
 *	2	-	+	.
 *	3	.	+	-
 *	4	.	-	+
 *	5	+	-	.
 *	6	-	.	+
 *	7	.	.	.
 */

// W3  W2  W1
//  5   5   5
static const uint32_t ccerLUT[8] = {
	0x555,
	0x505,
	0x055,
	0x550,
	0x550,
	0x055,
	0x505,
	0x000
};

static const uint32_t ccmr1LUT[8] = {
	CCMR1_OC1_LOW_ON | CCMR1_OC2_LOW_ON | CCMRx_OCxPE,
	CCMR1_OC1_PWM2 | CCMRx_OCxPE,
	CCMR1_OC1_LOW_ON | CCMR1_OC2_PWM2 | CCMRx_OCxPE,
	CCMR1_OC2_PWM2 | CCMRx_OCxPE,
	CCMR1_OC2_LOW_ON | CCMRx_OCxPE,
	CCMR1_OC1_PWM2 | CCMR1_OC2_LOW_ON | CCMRx_OCxPE,
	CCMR1_OC1_LOW_ON | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
};

static const uint32_t ccmr2LUT[8] = {
	CCMR2_OC3_LOW_ON | CCMRx_OCxPE,
	CCMR2_OC3_LOW_ON | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
	CCMR2_OC3_LOW_ON | CCMRx_OCxPE,
	CCMR2_OC3_PWM2 | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
	CCMR2_OC3_PWM2 | CCMRx_OCxPE,
	0 | CCMRx_OCxPE,
};

static binary_semaphore_t runSem;

static float min3(float a, float b, float c);
static void updateADC(Drive* pDrive);
static inline void updateHall(Drive* pDrive, uint16_t hallPos);
static void updateSpeed(Drive* pDrive);

/**
 * IRQs
 */
 // Encoder time measurement
IN_CCM void TIM2_IRQHandler()
{
	static IN_CCM_VAR uint16_t last = 0;

	register uint16_t now = TIM3->CNT;
	TIM3->CCR3 = now+64000;

	TIM2->SR = 0;

#ifdef MOT_RIGHT
	++pRear->encIndex;		// M4
	pRear->encIndex %= DRIVE_ENC_SAMPLES;
	pRear->encDtA[pRear->encIndex] = now-last;
#else
	++pFront->encIndex;		// M2
	pFront->encIndex %= DRIVE_ENC_SAMPLES;
	pFront->encDtA[pFront->encIndex] = now-last;
#endif

	last = now;
}

IN_CCM void TIM4_IRQHandler()
{
	static IN_CCM_VAR uint16_t last = 0;

	register uint16_t now = TIM3->CNT;
	TIM3->CCR4 = now+64000;

	TIM4->SR = 0;

#ifdef MOT_RIGHT
	++pFront->encIndex;		// M1
	pFront->encIndex %= DRIVE_ENC_SAMPLES;
	pFront->encDtA[pFront->encIndex] = now-last;
#else
	++pRear->encIndex;		// M3
	pRear->encIndex %= DRIVE_ENC_SAMPLES;
	pRear->encDtA[pRear->encIndex] = now-last;
#endif

	last = now;
}

IN_CCM void TIM3_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	register uint16_t sr = TIM3->SR;
	TIM3->SR = 0;
	uint16_t now = TIM3->CNT;

	if(sr & TIM_SR_CC1IF)
	{
		TIM3->CCR1 = now+3999;

		chSysLockFromISR();
		chBSemSignalI(&runSem);
		chSysUnlockFromISR();
	}

	if(sr & TIM_SR_CC3IF)
		TIM2->EGR = TIM_EGR_CC1G;

	if(sr & TIM_SR_CC4IF)
		TIM4->EGR = TIM_EGR_CC1G;

	CH_IRQ_EPILOGUE();
}
 
IN_CCM void TIM8_UP_IRQHandler()
{
	TIM8->SR = 0;

#ifdef MOT_RIGHT
	updateADC(pRear);	// M4
#else
	updateADC(pFront);	// M2
#endif
}

IN_CCM void TIM1_UP_TIM16_IRQHandler()
{
	TIM1->SR = 0;

#ifdef MOT_RIGHT
	updateADC(pFront);	// M1
#else
	updateADC(pRear);	// M3
#endif
}

#define M13HallMask (EXTI_PR_PR10 | EXTI_PR_PR11 | EXTI_PR_PR12)

IN_CCM void EXTI3_IRQHandler()
{
	EXTI->PR = EXTI_PR_PR3;

	uint16_t pos24 = ((GPIOA->IDR & GPIO_PIN_3) >> 3) | ((GPIOC->IDR & (GPIO_PIN_14 | GPIO_PIN_15)) >> 13);

#ifdef MOT_RIGHT
	updateHall(pRear, pos24);
#else
	updateHall(pFront, pos24);
#endif
}

IN_CCM void EXTI15_10_IRQHandler()
{
	uint32_t pr = EXTI->PR;
	EXTI->PR = M13HallMask | EXTI_PR_PR14 | EXTI_PR_PR15;

	if(pr & M13HallMask)
	{
		uint16_t pos13 = (GPIOB->IDR & M13HallMask) >> 10;

#ifdef MOT_RIGHT
		updateHall(pFront, pos13);
#else
		updateHall(pRear, pos13);
#endif
	}

	if(pr & (EXTI_PR_PR14 | EXTI_PR_PR15))
	{
		uint16_t pos24 = ((GPIOA->IDR & GPIO_PIN_3) >> 3) | ((GPIOC->IDR & (GPIO_PIN_14 | GPIO_PIN_15)) >> 13);

#ifdef MOT_RIGHT
		updateHall(pRear, pos24);
#else
		updateHall(pFront, pos24);
#endif
	}
}

/**
 * PRIVATE functions
 */
static void initHw()
{
	GPIOInitData gpioInit;

	// data structures
	pFront->bat = 12.0f;
	pRear->bat = 12.0f;

	// ### DMA configuration (ADC)
	// peripheral-to-memory, half-words, medium priority, memory increment, circular mode
	DMA1_Channel1->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA1_Channel1->CNDTR = DRIVE_ADC_SAMPLES;
	DMA1_Channel1->CPAR = (uint32_t)&ADC1->DR;
#ifdef MOT_RIGHT
	DMA1_Channel1->CMAR = (uint32_t)pRear->adc[0];
#else
	DMA1_Channel1->CMAR = (uint32_t)pFront->adc[0];
#endif

	DMA2_Channel1->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel1->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel1->CPAR = (uint32_t)&ADC2->DR;
#ifdef MOT_RIGHT
	DMA2_Channel1->CMAR = (uint32_t)pRear->adc[1];
#else
	DMA2_Channel1->CMAR = (uint32_t)pFront->adc[1];
#endif

	DMA2_Channel5->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel5->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel5->CPAR = (uint32_t)&ADC3->DR;
#ifdef MOT_RIGHT
	DMA2_Channel5->CMAR = (uint32_t)pFront->adc[0];
#else
	DMA2_Channel5->CMAR = (uint32_t)pRear->adc[0];
#endif

	DMA2_Channel2->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel2->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel2->CPAR = (uint32_t)&ADC4->DR;
#ifdef MOT_RIGHT
	DMA2_Channel2->CMAR = (uint32_t)pFront->adc[1];
#else
	DMA2_Channel2->CMAR = (uint32_t)pRear->adc[1];
#endif

	// ### ADC
	RCC->AHBENR |= RCC_AHBENR_ADC12EN | RCC_AHBENR_ADC34EN;

	gpioInit.mode = GPIO_MODE_ANALOG;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_2, &gpioInit); // A2
	GPIOInit(GPIOB, GPIO_PIN_2 | GPIO_PIN_13 | GPIO_PIN_14, &gpioInit); // B2, B13, B14

	// independent, synchronous clock from AHB
	ADC1_2->CCR = ADC12_CCR_CKMODE_0;
	ADC3_4->CCR = ADC34_CCR_CKMODE_0;

	// enable voltage regulator for each ADC
	ADC1->CR = ADC_CR_ADVREGEN_0;
	ADC2->CR = ADC_CR_ADVREGEN_0;
	ADC3->CR = ADC_CR_ADVREGEN_0;
	ADC4->CR = ADC_CR_ADVREGEN_0;
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	// start single-ended calibration
	ADC1->CR |= ADC_CR_ADCAL;
	ADC2->CR |= ADC_CR_ADCAL;
	ADC3->CR |= ADC_CR_ADCAL;
	ADC4->CR |= ADC_CR_ADCAL;
	chThdSleep(MS2ST(4));	// calibration delay

#ifdef MOT_RIGHT
	pRear->adcCal[0]  = ADC1->CALFACT;
	pRear->adcCal[1]  = ADC2->CALFACT;
	pFront->adcCal[0] = ADC3->CALFACT;
	pFront->adcCal[1] = ADC4->CALFACT;
#else
	pFront->adcCal[0] = ADC1->CALFACT;
	pFront->adcCal[1] = ADC2->CALFACT;
	pRear->adcCal[0]  = ADC3->CALFACT;
	pRear->adcCal[1]  = ADC4->CALFACT;
#endif

	// continuous conversion, 12Bit, trigger TIM8_TRGO2
	ADC1->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (8 << 6) | ADC_CFGR_DMACFG;
	ADC1->SQR1 = DRIVE_ADC_SAMPLES-1;

	ADC2->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (8 << 6) | ADC_CFGR_DMACFG;
	ADC2->SQR1 = DRIVE_ADC_SAMPLES-1;

	// continuous conversion, 12Bit, trigger TIM1_TRGO2
	ADC3->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC3->SQR1 = DRIVE_ADC_SAMPLES-1;

	ADC4->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC4->SQR1 = DRIVE_ADC_SAMPLES-1;

	for(uint8_t i = 1; i <= DRIVE_ADC_SAMPLES; i++)
	{
		ADCRegularChannelConfig(ADC1,  3, i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC2, 12, i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC3,  5, i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC4,  4, i, ADC_SampleTime_181Cycles5);
	}

	// configure offsets
	ADC1->OFR1 = ADC_OFR1_OFFSET1_EN | ( 3 << 26) | 1706;
	ADC2->OFR1 = ADC_OFR1_OFFSET1_EN | (12 << 26) | 1706;
	ADC3->OFR1 = ADC_OFR1_OFFSET1_EN | ( 5 << 26) | 1706;
	ADC4->OFR1 = ADC_OFR1_OFFSET1_EN | ( 4 << 26) | 1706;

	// Enable the ADC DMA
	ADC1->CFGR |= ADC_CFGR_DMAEN;
	ADC2->CFGR |= ADC_CFGR_DMAEN;
	ADC3->CFGR |= ADC_CFGR_DMAEN;
	ADC4->CFGR |= ADC_CFGR_DMAEN;

	ADC1->CR |= ADC_CR_ADEN;
	ADC2->CR |= ADC_CR_ADEN;
	ADC3->CR |= ADC_CR_ADEN;
	ADC4->CR |= ADC_CR_ADEN;
	chThdSleep(MS2ST(4));	// ADC ready delay

	// Enable the DMA channel
	DMA1_Channel1->CCR |= DMA_CCR_EN;
	DMA2_Channel1->CCR |= DMA_CCR_EN;
	DMA2_Channel5->CCR |= DMA_CCR_EN;
	DMA2_Channel2->CCR |= DMA_CCR_EN;

	// ### Encoder
	RCC->APB1ENR |= RCC_APB1ENR_TIM2EN | RCC_APB1ENR_TIM4EN;

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 1;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_10MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_0 | GPIO_PIN_1, &gpioInit);	// TIM2_CH1, TIM2_CH2

	gpioInit.alternate = 10;
	GPIOInit(GPIOA, GPIO_PIN_11, &gpioInit);				// TIM4_CH1

	gpioInit.alternate = 2;
	GPIOInit(GPIOB, GPIO_PIN_7, &gpioInit);					// TIM4_CH2

	//Timer Init - Encoder inputs
	TIM2->CR1 = 0;
	TIM2->CR2 = 0;
	TIM2->PSC = 0;
	TIM2->ARR = 0xFFFF;
	TIM2->RCR = 0;
	// IC1 mapped to TI1, IC2 mapped to TI2, filter = 0x0F on both, no prescaler
	TIM2->CCMR1 = TIM_CCMR1_CC1S_0 | (0x07 << 4) | TIM_CCMR1_CC2S_0 | (0x07 << 12);
	TIM2->CCER = TIM_CCER_CC1E | TIM_CCER_CC2E;	// enable input capture, rising polarity
	TIM2->DIER = TIM_DIER_CC1IE;
	TIM2->SMCR = 3;	// encoder mode 3, count all flanks
	TIM2->CR1 |= TIM_CR1_CEN;

	TIM4->CR1 = 0;
	TIM4->CR2 = 0;
	TIM4->PSC = 0;
	TIM4->ARR = 0xFFFF;
	TIM4->RCR = 0;
	// IC1 mapped to TI1, IC2 mapped to TI2, filter = 0x0F on both, no prescaler
	TIM4->CCMR1 = TIM_CCMR1_CC1S_0 | (0x07 << 4) | TIM_CCMR1_CC2S_0 | (0x07 << 12);
	TIM4->CCER = TIM_CCER_CC1E | TIM_CCER_CC2E;	// enable input capture, rising polarity
	TIM4->DIER = TIM_DIER_CC1IE;
	TIM4->SMCR = 3;	// encoder mode 3, count all flanks
	TIM4->CR1 |= TIM_CR1_CEN;

	NVICEnableIRQ(TIM2_IRQn, 1);	// warning: this is out of kernel IRQ level range. Do not call any kernel functions from IRQ!
	NVICEnableIRQ(TIM4_IRQn, 1);

	// ### Encoder Velocity Measurement Timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;

	TIM3->CR1 = 0;
	TIM3->CR2 = 0;
	TIM3->PSC = 17;	// = 250ns = 4MHz
	TIM3->ARR = 0xFFFF;	// 65535
	TIM3->RCR = 0;

	TIM3->CCMR1 = 0;
	TIM3->CCR1 = 4000;
	TIM3->CCR3 = 64000;
	TIM3->CCR4 = 64000;

	TIM3->CR1 |= TIM_CR1_CEN;

	TIM3->DIER = TIM_DIER_CC1IE | TIM_DIER_CC3IE | TIM_DIER_CC4IE;

	NVICEnableIRQ(TIM3_IRQn, IRQL_SPEED_UPDATE);

	// ### Motors
	RCC->APB2ENR |= RCC_APB2ENR_TIM1EN | RCC_APB2ENR_TIM8EN;
	RCC->CFGR3 |= RCC_CFGR3_TIM1SW | RCC_CFGR3_TIM8SW;	// TIM1 + TIM8 clocks now running at 144MHz

	// Timer config
	// TIM1
	TIM1->CR1 = TIM_CR1_CMS_0;	// center-aligned mode 1
	TIM1->PSC = 0;
	TIM1->ARR = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	TIM1->RCR = 1;				// update event only every second count (e.g. downcounting) = 20kHz

	// PWM mode 2, preload enable on CCR(1-3) registers
	TIM1->CCMR1 = TIM_CCMR1_OC1PE | (7 << 4) | TIM_CCMR1_OC2PE | (7 << 12);
	TIM1->CCMR2 = TIM_CCMR2_OC3PE | (7 << 4) | (7 << 12);
	TIM1->CCER = TIM_CCER_CC1E | TIM_CCER_CC1NE | TIM_CCER_CC2E | TIM_CCER_CC2NE | TIM_CCER_CC3E | TIM_CCER_CC3NE | TIM_CCER_CC4E;
	TIM1->CR2 = TIM_CR2_OIS1 | TIM_CR2_OIS2 | TIM_CR2_OIS3 | (7 << 20);	// output idle state = high, OC4 on TRGO2 (connected to ADC)
	TIM1->CCR1 = 0;
	TIM1->CCR2 = 0;
	TIM1->CCR3 = 0;
	TIM1->CCR4 = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM1->EGR = TIM_EGR_UG;

	TIM1->BDTR = TIM_BDTR_AOE | TIM_BDTR_BKP | TIM_BDTR_OSSR;
	TIM1->DIER = TIM_DIER_UIE;

	TIM1->CR1 |= TIM_CR1_CEN;

	ADC1->CR |= ADC_CR_ADSTART;
	ADC2->CR |= ADC_CR_ADSTART;

	TIM1->BDTR |= TIM_BDTR_MOE;

	// TIM8
	TIM8->CR1 = TIM_CR1_CMS_0;	// center-aligned mode 1
	TIM8->PSC = 0;
	TIM8->ARR = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	TIM8->RCR = 1;				// update event only every second count (e.g. downcounting) = 20kHz

	// PWM mode 2, preload enable on CCR(1-3) registers
	TIM8->CCMR1 = TIM_CCMR1_OC1PE | (7 << 4) | TIM_CCMR1_OC2PE | (7 << 12);
	TIM8->CCMR2 = TIM_CCMR2_OC3PE | (7 << 4) | (7 << 12);
	TIM8->CCER = TIM_CCER_CC1E | TIM_CCER_CC1NE | TIM_CCER_CC2E | TIM_CCER_CC2NE | TIM_CCER_CC3E | TIM_CCER_CC3NE | TIM_CCER_CC4E;
	TIM8->CR2 = TIM_CR2_OIS1 | TIM_CR2_OIS2 | TIM_CR2_OIS3 | (7 << 20);	// output idle state = high, OC4 on TRGO2 (connected to ADC)
	TIM8->CCR1 = 0;
	TIM8->CCR2 = 0;
	TIM8->CCR3 = 0;
	TIM8->CCR4 = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM8->EGR = TIM_EGR_UG;

	TIM8->BDTR = TIM_BDTR_AOE | TIM_BDTR_BKP | TIM_BDTR_OSSR;
	TIM8->DIER = TIM_DIER_UIE;

	TIM8->CR1 |= TIM_CR1_CEN;

	ADC3->CR |= ADC_CR_ADSTART;
	ADC4->CR |= ADC_CR_ADSTART;

	TIM8->BDTR |= TIM_BDTR_MOE;

	// pin config - drive outputs
	GPIOA->ODR |= GPIO_PIN_7 | GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_12;
	GPIOB->ODR |= GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_6 | GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_15;
	GPIOC->ODR |= GPIO_PIN_13;

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 4;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_50MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_7, &gpioInit);
	GPIOInit(GPIOB, GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_15, &gpioInit);
	GPIOInit(GPIOC, GPIO_PIN_13, &gpioInit);

	gpioInit.alternate = 6;
	GPIOInit(GPIOA, GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_12, &gpioInit);

	gpioInit.alternate = 5;
	GPIOInit(GPIOB, GPIO_PIN_6, &gpioInit);

	gpioInit.alternate = 10;
	GPIOInit(GPIOB, GPIO_PIN_8 | GPIO_PIN_9, &gpioInit);

	// ### Hall Sensors
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING_FALLING;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_3, &gpioInit);
	GPIOInit(GPIOB, GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit);
	GPIOInit(GPIOC, GPIO_PIN_14 | GPIO_PIN_15, &gpioInit);

	NVICEnableIRQ(TIM8_UP_IRQn, 2);
	NVICEnableIRQ(TIM1_UP_TIM16_IRQn, 2);
	NVICEnableIRQ(EXTI3_IRQn, 1);
	NVICEnableIRQ(EXTI15_10_IRQn, 1);
}

IN_CCM static float min3(float a, float b, float c)
{
	if(a < b && a < c)
		return a;

	if(b < a && b < c)
		return b;

	return c;
}

typedef union _i32i16
{
	int32_t i32;
	int16_t i16[2];
} i32i16_t;

IN_CCM static inline void updateADC(Drive* pDrive)
{
	int16_t i016, i116;
	float32_t i0f, i1f;
	i32i16_t i0i, i1i;

	i0i.i32 = pDrive->adc32[0][0];
	i1i.i32 = pDrive->adc32[1][0];

	for(uint8_t i = 1; i < DRIVE_ADC_SAMPLES/2; i++)
	{
		i0i.i32 = __QADD16(i0i.i32, pDrive->adc32[0][i]);
		i1i.i32 = __QADD16(i1i.i32, pDrive->adc32[1][i]);
	}

	i016 = i0i.i16[0]+i0i.i16[1];
	i116 = i1i.i16[0]+i1i.i16[1];

	i0f = i016*(1.0f/DRIVE_ADC_SAMPLES);
	i1f = i116*(1.0f/DRIVE_ADC_SAMPLES);

	i0f = i0f*-MOT_ADC_TO_CURRENT;
	i1f = i1f*-MOT_ADC_TO_CURRENT;

	float a = i0f;
	float b = 0.57735026918963f*i0f + 1.154700538379252f*i1f;

	const float alpha = 0.995f;
	float rms = sqrtf(a*a + b*b);
	pDrive->rmsCurrent = (alpha*pDrive->rmsCurrent + (1.0f-alpha)*rms);

	pDrive->rmsCurrentTimestamp = SysTimeUSec();

	// call something on current update?
	// Rate: 20kHz, above kernel IRQ level
}

IN_CCM static void updateSpeed(Drive* pDrive)
{
	float vel;

	uint16_t dir = (pDrive->pTIMEncoder->CR1 & TIM_CR1_DIR);
	uint32_t encDt = 0;
	uint16_t steps = 0;
	for(uint8_t i = 0; i < DRIVE_ENC_SAMPLES; i++)
	{
		uint16_t dt = pDrive->encDtA[i];
		encDt += dt;
		if(dt < 64000)
			steps++;
	}

	if(encDt > steps*64000 || encDt < 80)
	{
		vel = 0;
	}
	else if(dir != pDrive->lastDir)
	{
		vel = 0;
		memset(pDrive->encDtA, 0xFF, DRIVE_ENC_SAMPLES*sizeof(uint16_t));
	}
	else
	{
		vel = (steps*1e6f)/(encDt*128);

		if(dir)
			vel *= -1;
	}

	pDrive->lastDir = dir;

	if(vel < 10000)
	{
		pDrive->curVelocity = vel;
		pDrive->curVelocityTimestamp = SysTimeUSec();
	}
}

IN_CCM static inline void updateHall(Drive* pDrive, uint16_t hallPos)
{
	pDrive->halPos = hallPos;

	if(pDrive->mode != DRIVE_MODE_BLOCK)
		return;
		
	if(pDrive->outLevel < 0)
	{
		hallPos ^= 0x7;
	}

	pDrive->pTIMMotor->CCER = ccerLUT[hallPos];
	
	pDrive->pTIMMotor->CCMR1 = ccmr1LUT[hallPos];
	pDrive->pTIMMotor->CCMR2 = ccmr2LUT[hallPos];
}

IN_CCM static void setCCR(Drive* pDrive, uint16_t a, uint16_t b, uint16_t c)
{
	pDrive->pTIMMotor->CCR1 = a;
	pDrive->pTIMMotor->CCR2 = b;
	pDrive->pTIMMotor->CCR3 = c;
}

void DriveTask(void* params)
{
	(void)params;

	chRegSetThreadName("Drive");

	while(1)
	{
		chBSemWait(&runSem);

		for(uint8_t m = 0; m < 2; m++)
		{
			Drive* pDrive = (Drive*)&drive[m];

			updateSpeed(pDrive);

			pDrive->estCurrent = 1.0f/MOTOR_RESISTANCE*(pDrive->setVoltage-MOTOR_KE*pDrive->curVelocity*2*M_PI);

			if(pDrive->ctrlMode == DRIVE_CTRL_MODE_PID_SPEED)
			{
				// control
				PIDSetLimits(&pDrive->speed.pid, -pDrive->bat, pDrive->bat);
				pDrive->speed.pid.setpoint = pDrive->speed.setVelocity;
				PIDUpdate(&pDrive->speed.pid, pDrive->curVelocity);

				DriveSetVoltage(pDrive, pDrive->speed.pid.output);
			}
		}
	}
}

/**
 * PUBLIC functions
 */
void DriveInit()
{
	chBSemObjectInit(&runSem, 1);

	for(uint8_t i = 0; i < 2; i++)
	{
		Drive* pDrive = &drive[i];

		pDrive->slopeLimited = 1;

		PIDParams speedPIDParams;
		speedPIDParams.kp = 0.5f;
		speedPIDParams.ki = 50.0f;
		speedPIDParams.kd = 0.0035f;
		speedPIDParams.outputMin = -10.0f;	// [V]
		speedPIDParams.outputMax = 10.0f;	// [V]

		PIDInit(&pDrive->speed.pid, &speedPIDParams, 0.001f, 1);
	}

	initHw();

	DriveSetMode(pFront, DRIVE_MODE_BLOCK);
	DriveSetMode(pRear, DRIVE_MODE_BLOCK);

	DriveSetCtrlMode(pFront, DRIVE_CTRL_MODE_PID_SPEED);
	DriveSetCtrlMode(pRear, DRIVE_CTRL_MODE_PID_SPEED);
}

void DriveSetMode(Drive* pDrive, uint8_t mode)
{
	if(mode == pDrive->mode)
		return;

	pDrive->mode = mode;

	switch(mode)
	{
		case DRIVE_MODE_DISABLE:
		{
			pDrive->pTIMMotor->CCER = ccerLUT[7];
			pDrive->outLevel = 0;
		}
		break;
		case DRIVE_MODE_BLOCK:
		{
			// trigger hall IRQ to get a valid hall position
			EXTI->SWIER |= (EXTI_SWIER_SWIER10 | EXTI_SWIER_SWIER14);
		}
		break;
		case DRIVE_MODE_SVPWM:
		{
			pDrive->outLevel = 0;
			pDrive->pTIMMotor->CCR1 = 0;
			pDrive->pTIMMotor->CCR2 = 0;
			pDrive->pTIMMotor->CCR3 = 0;
			pDrive->pTIMMotor->CCER = 0x555;
			pDrive->pTIMMotor->CCMR1 = CCMR1_OC1_PWM2 | CCMR1_OC2_PWM2 | CCMRx_OCxPE;
			pDrive->pTIMMotor->CCMR2 = CCMR2_OC3_PWM2 | CCMRx_OCxPE;
		}
		break;
		case DRIVE_MODE_TEST1:
		{
			// W1+, W2-
			pDrive->pTIMMotor->CCER = ccerLUT[5];
			pDrive->pTIMMotor->CCMR1 = ccmr1LUT[5];
			pDrive->pTIMMotor->CCMR2 = ccmr2LUT[5];
		}
		break;
		case DRIVE_MODE_TEST2:
		{
			// W2+, W3-
			pDrive->pTIMMotor->CCER = ccerLUT[3];
			pDrive->pTIMMotor->CCMR1 = ccmr1LUT[3];
			pDrive->pTIMMotor->CCMR2 = ccmr2LUT[3];
		}
		break;
		case DRIVE_MODE_TEST3:
		{
			// W3+, W1-
			pDrive->pTIMMotor->CCER = ccerLUT[6];
			pDrive->pTIMMotor->CCMR1 = ccmr1LUT[6];
			pDrive->pTIMMotor->CCMR2 = ccmr2LUT[6];
		}
		break;
	}
}

void DriveSetCtrlMode(Drive* pDrive, uint8_t mode)
{
	if(pDrive->ctrlMode == mode)
		return;

	switch(mode)
	{
		case DRIVE_CTRL_MODE_NONE:
		{
			PIDEnable(&pDrive->speed.pid, 0);
		}
		break;
		case DRIVE_CTRL_MODE_PID_SPEED:
		{
			PIDEnable(&pDrive->speed.pid, 1);
		}
		break;
		default:
			return;
	}

	pDrive->ctrlMode = mode;
}

#define MAX_PWM_CHANGE 50

IN_CCM void DriveSetVoltage(Drive* pDrive, float voltage)
{
	pDrive->setVoltage = voltage;
	pDrive->setVoltageTimestamp = SysTimeUSec();

	int32_t newOutLevel = (int32_t)(pDrive->setVoltage/pDrive->bat*((float)PWM_MAX_VAL));
	if(newOutLevel > PWM_MAX_VAL)
		newOutLevel = PWM_MAX_VAL;

	if(newOutLevel < -PWM_MAX_VAL)
		newOutLevel = -PWM_MAX_VAL;

	if(pDrive->slopeLimited)
	{
		int32_t pwmChange = newOutLevel-pDrive->outLevel;
		if(pwmChange > MAX_PWM_CHANGE)
			pwmChange = MAX_PWM_CHANGE;

		if(pwmChange < -MAX_PWM_CHANGE)
			pwmChange = -MAX_PWM_CHANGE;

		pDrive->outLevel += pwmChange;
		newOutLevel = pDrive->outLevel;
	}
	else
	{
		pDrive->outLevel = newOutLevel;
	}

	updateHall(pDrive, pDrive->halPos);

	if(newOutLevel < 0)
		newOutLevel *= -1;

	pDrive->pTIMMotor->CCR1 = newOutLevel;
	pDrive->pTIMMotor->CCR2 = newOutLevel;
	pDrive->pTIMMotor->CCR3 = newOutLevel;
}

void DriveSpeedSetPIDParams(Drive* pDrive, float p, float i, float d)
{
	PIDSetParams(&pDrive->speed.pid, p, i, d);
}

void DriveSpeedSetVelocity(Drive* pDrive, float vel)
{
	pDrive->speed.setVelocity = vel;
}

void DriveSetPhaseVoltagesSVPWM(Drive* pDrive, float Va, float Vb, float Vc)
{
	DriveSetMode(pDrive, DRIVE_MODE_SVPWM);

	float Vd = pDrive->bat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float Vcom = -min3(Va, Vb, Vc);

	float da = 1/Vd*(Va+Vcom);
	float db = 1/Vd*(Vb+Vcom);
	float dc = 1/Vd*(Vc+Vcom);

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(pDrive, da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

void DriveSetAlphaBeta(Drive* pDrive, float Valpha, float Vbeta)
{
	float Va = Valpha;
	float Vb = -0.5f*Valpha + 0.86602540378444f*Vbeta;
	float Vc = -0.5f*Valpha - 0.86602540378444f*Vbeta;

	DriveSetPhaseVoltagesSVPWM(pDrive, Va, Vb, Vc);
}

void DriveSetVoltageAngle(Drive* pDrive, float Vref, float angle)
{
	float sin, cos;
	while(angle > 180.0f)
		angle -= 360.0f;

	arm_sin_cos_f32(angle, &sin, &cos);

	DriveSetAlphaBeta(pDrive, cos*Vref, sin*Vref);
}
