/*
 * log_file.c
 *
 *  Created on: 24.08.2014
 *      Author: AndreR
 */

#include "log_file.h"

#include "ch.h"
#include "util/intercom.h"
#include "intercom_constants.h"
#include "util/sys_time.h"
#include "util/log.h"
#include "log_msgs.h"
#include "struct_ids.h"
#include "util/cobs.h"
#include <string.h>
#include <stdio.h>

#define EVENT_CLOSE			0x10000
#define EVENT_OPEN_NAMED	0x20000
#define EVENT_OPEN_AUTO		0x40000

LogFile logFile;

static void writeFile(uint32_t bufIndex)
{
	if(chMBPost(&logFile.eventQueue, bufIndex, MS2ST(1000)) != MSG_OK)
	{
		LogErrorC("Post event failed", bufIndex);
		FifoBlockFreeBuffer(&logFile.writeFifo, bufIndex);
	}
}

static void writeFileDirect(uint32_t bufIndex)
{
	FRESULT fresult;
	uint8_t* pData;
	uint32_t dataSize;
	UINT bytesWritten;

	if(FifoBlockGetData(&logFile.writeFifo, bufIndex, &pData, &dataSize) != 0)
	{
		LogErrorC("Get FIFO block failed", bufIndex);
		return;
	}

	fresult = f_write(&logFile.file, pData, dataSize, &bytesWritten);
	if(fresult != FR_OK)
	{
		logFile.writeFail = 1;
		LogErrorC("f_write failed", fresult);
	}
	if(dataSize != bytesWritten)
	{
		LogErrorC("dataSize != bytesWritten", bytesWritten);
	}

	logFile.fileSize += bytesWritten;

	FifoBlockFreeBuffer(&logFile.writeFifo, bufIndex);
}

static void logfileCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	uint16_t bytesWritten = 0;

	if(pHeader->command == IC_LOGFILE_MESSAGE)
	{
		while(bytesWritten + sizeof(LogChunkHeader) < dataLength)
		{
			if(((LogChunkHeader*)pData)->length == 0)
			{
				LogErrorC("zero length logfile entry", ((LogChunkHeader*)pData)->type);
				break;
			}

			LogFileWrite(pData);

			LogChunkHeader* pHeader = (LogChunkHeader*)pData;
			pData += pHeader->length;
			bytesWritten += pHeader->length;
		}
	}
}

void LogFileInit(uint8_t* pWriteBuf, uint32_t numChunks, uint32_t chunkSize)
{
	logFile.ready = 0;
	chMBObjectInit(&logFile.eventQueue, logFile.eventQueueData, LOG_FILE_QUEUE_SIZE);

	chMtxObjectInit(&logFile.encodeMtx);

	FifoBlockInit(&logFile.writeFifo, numChunks, chunkSize, pWriteBuf);
	logFile.writeFifo.fullCb = &writeFile;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &logfileCallback;
	client.section = INTERCOM_SECTION_LOGFILE;
	IntercomAddClient(client);
}

void LogFileOpen(const char* pName)
{
	if(logFile.ready)
		LogFileClose();

	if(pName)
	{
		snprintf(logFile.logFilename, 64, "%s.dat", pName);
		if(chMBPost(&logFile.eventQueue, EVENT_OPEN_NAMED, MS2ST(500)) != MSG_OK)
		{
			LogError("post failed");
		}
	}
	else
	{
		if(chMBPost(&logFile.eventQueue, EVENT_OPEN_AUTO, MS2ST(500)) != MSG_OK)
		{
			LogError("post failed");
		}
	}
}

void LogFileClose()
{
	if(chMBPost(&logFile.eventQueue, EVENT_CLOSE, MS2ST(500)) != MSG_OK)
	{
		LogError("post failed");
	}
}

static void encodeAndStore(void* pLogMsg)
{
	uint32_t bytesWritten;
	LogChunkHeader* pHeader = (LogChunkHeader*)pLogMsg;

	chMtxLock(&logFile.encodeMtx);

	if(COBSEncode(pLogMsg, pHeader->length, logFile.cobsEncodeBuf, LOG_FILE_WORKING_BUF_SIZE-1, &bytesWritten) != 0)
	{
		LogError("Encoding log msg failed");
		chMtxUnlock(&logFile.encodeMtx);
		return;
	}

	logFile.cobsEncodeBuf[bytesWritten] = 0;

	logFile.bytesPushed += bytesWritten;

	FifoBlockPush(&logFile.writeFifo, logFile.cobsEncodeBuf, bytesWritten+1, 1);

	chMtxUnlock(&logFile.encodeMtx);
}

void LogFileWrite(void* pLogMsg)
{
	if(logFile.ready)
		encodeAndStore(pLogMsg);
}

int16_t LogFilePackConfigFile(const ConfigFile* pFile, uint8_t* pOut, uint32_t outSize, uint32_t* pBytesWritten)
{
	uint32_t msgSize = pFile->size + sizeof(LogEntryHeader);

	*pBytesWritten = msgSize;

	if(outSize < msgSize)
	{
		LogErrorC("out buffer too small", msgSize);
		return 1;
	}

	LogEntryHeader* pHeader = (LogEntryHeader*)pOut;
	pHeader->type = pFile->pDesc->cfgId;
	pHeader->length = pFile->size + sizeof(LogEntryHeader);
	pHeader->timestamp = SysTimeUSec();

	pOut += sizeof(LogEntryHeader);

	memcpy(pOut, pFile->pData, pFile->size);

	return 0;
}

int16_t LogFilePackConfigDesc(const ConfigFileDesc* pCfg, uint8_t* pOut, uint32_t outSize, uint32_t* pBytesWritten)
{
	char* pReplace;
	uint32_t msgSize = 0;

	LogChunkHeader cHeader;
	cHeader.type = SID_MESSAGE_DESC;
	cHeader.length = sizeof(uint16_t)+sizeof(uint16_t)+strlen(pCfg->pName)+1;	// msg ID, numElements, name[]

	msgSize += cHeader.length + sizeof(LogChunkHeader);

	for(uint16_t el = 0; el < pCfg->numElements; el++)
	{
		const ElementDesc* pElDesc = &pCfg->elements[el];

		cHeader.type = SID_ELEMENT_DESC;
		cHeader.length = sizeof(uint8_t) + strlen(pElDesc->pUnit)		// type and unit
				+ strlen(pElDesc->pName) + 3;	// string lengths

		msgSize += cHeader.length + sizeof(LogChunkHeader);
	}

	*pBytesWritten = msgSize;

	if(outSize < msgSize)
	{
		LogErrorC("out buffer too small", msgSize);
		return 1;
	}

	cHeader.type = SID_MESSAGE_DESC;
	cHeader.length = sizeof(LogChunkHeader)+sizeof(uint16_t)+sizeof(uint16_t)+strlen(pCfg->pName)+1;	// msg ID, numElements, name[]

	memcpy(pOut, &cHeader, sizeof(LogChunkHeader));
	pOut += sizeof(LogChunkHeader);
	memcpy(pOut, &pCfg->cfgId, sizeof(uint16_t));
	pOut += sizeof(uint16_t);
	memcpy(pOut, &pCfg->numElements, sizeof(uint16_t));
	pOut += sizeof(uint16_t);
	memcpy(pOut, pCfg->pName, strlen(pCfg->pName)+1);
	while((pReplace = strchr((char*)pOut, '/')) != 0)
		*pReplace = '_';
	pOut += strlen(pCfg->pName)+1;

	for(uint16_t el = 0; el < pCfg->numElements; el++)
	{
		const ElementDesc* pElDesc = &pCfg->elements[el];

		cHeader.type = SID_ELEMENT_DESC;
		cHeader.length = sizeof(LogChunkHeader) + sizeof(uint8_t) + strlen(pElDesc->pUnit)	// type and unit
				+ strlen(pElDesc->pName) + 3;	// string lengths

		memcpy(pOut, &cHeader, sizeof(LogChunkHeader));
		pOut += sizeof(LogChunkHeader);
		memcpy(pOut, &pElDesc->type, sizeof(uint8_t));
		pOut += sizeof(uint8_t);
		memcpy(pOut, pElDesc->pName, strlen(pElDesc->pName)+1);
		while((pReplace = strchr((char*)pOut, '/')) != 0)
			*pReplace = '_';
		pOut += strlen(pElDesc->pName)+1;
		memcpy(pOut, pElDesc->pUnit, strlen(pElDesc->pUnit)+1);
		pOut += strlen(pElDesc->pUnit)+1;
		*pOut = 0;
		pOut += 1;
	}

	return 0;
}

static void writeHeader()
{
	LogChunkHeader* pHeader = (LogChunkHeader*)logFile.initBuf;

	logFile.writeFifo.fullCb = &writeFileDirect;

	logFile.fileSize = 0;

	// write message descs
	for(uint32_t msg = 0; msg < sizeof(logMessageDescriptions)/sizeof(logMessageDescriptions[0]); msg++)
	{
		const LogMessageDesc* pMsgDesc = &logMessageDescriptions[msg];

		// now write the message
		pHeader->type = SID_MESSAGE_DESC;
		pHeader->length = sizeof(LogChunkHeader)+sizeof(uint16_t)+sizeof(uint16_t)+strlen(pMsgDesc->pName)+1;	// msg ID, numElements, name[]

		uint8_t* pWrite = logFile.initBuf+sizeof(LogChunkHeader);

		memcpy(pWrite, &pMsgDesc->msgId, sizeof(uint16_t));
		pWrite += sizeof(uint16_t);
		memcpy(pWrite, &pMsgDesc->numElements, sizeof(uint16_t));
		pWrite += sizeof(uint16_t);
		memcpy(pWrite, pMsgDesc->pName, strlen(pMsgDesc->pName)+1);

		encodeAndStore(pHeader);

		for(uint16_t el = 0; el < pMsgDesc->numElements; el++)
		{
			const ElementDesc* pElDesc = &pMsgDesc->elements[el];

			pHeader->type = SID_ELEMENT_DESC;
			pHeader->length = sizeof(LogChunkHeader)+ sizeof(uint8_t) + strlen(pElDesc->pUnit)	// type and unit
					+ strlen(pElDesc->pName) + strlen(pElDesc->pDesc) + 3;	// string lengths

			pWrite = logFile.initBuf+sizeof(LogChunkHeader);

			memcpy(pWrite, &pElDesc->type, sizeof(uint8_t));
			pWrite += sizeof(uint8_t);
			memcpy(pWrite, pElDesc->pName, strlen(pElDesc->pName)+1);
			pWrite += strlen(pElDesc->pName)+1;
			memcpy(pWrite, pElDesc->pUnit, strlen(pElDesc->pUnit)+1);
			pWrite += strlen(pElDesc->pUnit)+1;
			memcpy(pWrite, pElDesc->pDesc, strlen(pElDesc->pDesc)+1);

			encodeAndStore(pHeader);
		}
	}

	FifoBlockFlush(&logFile.writeFifo);

	logFile.writeFifo.fullCb = &writeFile;

	logFile.event.eventType = LOGFILE_EVENT_START;
	logFile.event.pFilename = logFile.logFilename;
	logFile.event.fileSize = 0;
	logFile.event.bytesPerSecIn = 0;
	logFile.event.writeSpeed = 0;

	logFile.ready = 1;

	if(logFile.eventCb)
		(*logFile.eventCb)(&logFile.event);
}

void LogFileTask(void* params)
{
	(void)params;

	msg_t event;
	uint8_t* pData;
	uint32_t dataSize;
	UINT bytesWritten;
	FRESULT fresult;

	chRegSetThreadName("LogFile");

	while(1)
	{
		if(chVTTimeElapsedSinceX(logFile.lastPushedTime) >= MS2ST(1000))
		{
			logFile.lastPushedTime = chVTGetSystemTimeX();

			logFile.event.eventType = LOGFILE_EVENT_UPDATE;
			logFile.event.bytesPerSecIn = logFile.bytesPushed;
			logFile.bytesPushed = 0;

			if(logFile.eventCb && logFile.ready)
				(*logFile.eventCb)(&logFile.event);
		}

		if(chMBFetch(&logFile.eventQueue, &event, MS2ST(100)) != MSG_OK)
		{
			if(logFile.writeFail)
			{
				LogFileClose();
			}

			continue;
		}

		switch(event)
		{
			case EVENT_OPEN_AUTO:
			{
				logFile.writeFail = 0;

				unsigned int id = 0;

				do
				{
					snprintf(logFile.logFilename, 64, "log%u.dat", id++);
					fresult = f_open(&logFile.file, logFile.logFilename, FA_WRITE | FA_CREATE_NEW);
				} while(fresult == FR_EXIST);

				if(fresult != FR_OK)
				{
					LogErrorC("f_open failed", fresult);
					break;
				}

				writeHeader();
			}
			break;
			case EVENT_OPEN_NAMED:
			{
				logFile.writeFail = 0;

				fresult = f_open(&logFile.file, logFile.logFilename, FA_WRITE | FA_CREATE_ALWAYS);

				if(fresult != FR_OK)
				{
					LogErrorC("f_open failed", fresult);
					break;
				}

				writeHeader();
			}
			break;
			case EVENT_CLOSE:
			{
				if(logFile.ready == 0)
					break;

				logFile.ready = 0;

				logFile.writeFifo.fullCb = &writeFileDirect;
				FifoBlockFlush(&logFile.writeFifo);
				logFile.writeFifo.fullCb = &writeFile;

				f_close(&logFile.file);
				logFile.file.fs = 0;

				logFile.event.eventType = LOGFILE_EVENT_CLOSE;
				logFile.event.fileSize = 0;
				logFile.event.bytesPerSecIn = 0;
				logFile.event.writeSpeed = 0;

				if(logFile.eventCb)
					(*logFile.eventCb)(&logFile.event);

				logFile.writeFail = 0;
			}
			break;
			default:
			{
				if(FifoBlockGetData(&logFile.writeFifo, event, &pData, &dataSize) == 0)
				{
					uint32_t start = SysTimeUSec();
					fresult = f_write(&logFile.file, pData, dataSize, &bytesWritten);
					f_sync(&logFile.file);
					uint32_t writeTime = SysTimeUSec()-start;

					if(fresult == FR_OK)
					{
						logFile.event.writeSpeed = ((1e9/writeTime)*bytesWritten)/1000;
						logFile.event.fileSize += bytesWritten;
					}
					else
					{
						logFile.writeFail = 1;
					}

					if(dataSize != bytesWritten)
					{
						LogErrorC("dataSize != bytesWritten", bytesWritten);
					}

					logFile.fileSize += bytesWritten;

					FifoBlockFreeBuffer(&logFile.writeFifo, event);
				}
			}
			break;
		}
	}
}

/**
General chunk header:
uint16_t chunkType	// MSB reserved for control chunks
uint16_t chunkLength

0x8000	message description
0x8001	element description
0x8002	unit description

File Order:
unit descriptions
message descriptions
- element desc
- ...
...
 */
